/*
 *
 *                  Copyright 1991,1992 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN

#define MAX_BUF_LEN    100

static LPVOID _cfw_hash_names        = NIL;  /* pointer to the hash table containing */
                                           /* all 'CFW' keyword names. */

static LPVOID _cfw_hash_templates    = NIL;  /* pointer to the hash table containing */
                                           /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _cfw_hash_templates.
 */
typedef struct _LanguageTemplate {         
	LPSTR abbrev;
	LPSTR template;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR CFW_INDENT_STR = "\377THEN";

/*
 * _cfw_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _cfw_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _cfw_templates[] = {
	{ "s|sp|speak",	"SPEAK & AS PROMPT" },
	{ "i|if",		"IF & = \n\tTHEN GO" },
	{ "m|menu",		"MENU\n\tON CN$TIMEOUT GOTO &\nON CN$INVALID GOSUB \nON [] GOTO \nON [] GOTO \nON [] GOTO \nGOTO " },
	{ "c|ch|char",	"CHAR:" },
	{ "n|num",		"NUM:" },
//	{ "\"",			""****************************************\n"&\n"****************************************\n" },
	{ "sub",		"STARTSUB &\n\nENDSUB" },
	{ "ca|cap",		"CAPTURE & AT  ON " },
	{ "p|pl|place",	"PLACE & AT  ON " },
	{ NIL,       NIL },
};

/* Define the list of 'CFW' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _cfw_names[] = {
	"ALL",        "APPEND",       "ARRAY",        "AS",           "AT",           "CALL",       
	"CAPTURE",    "CHAR",         "CHARS",        "CLEAR",        "CL$CALLINGID", "CL$CALLEDID",
	"CL$DIGIT",   "CL$PORT",      "CL$ESCODE",    "CN$CLABANDON", "CN$FAIL",      "CN$HOSTFAIL",
	"CN$INVALID", "CN$INVLIMIT",  "CN$LOCKED",    "CN$TIMEOUT",   "CN$TMOLIMIT",  "CONCAT",     
	"CONNECT",    "DAY",          "CB$RECORDNUM", "DEFAULT",      "DESELECT",     "DIGITS",     
	"DTMF",       "EDIT",         "ENDFLOW",      "ENDSUB",       "ERASE",        "FIND",       
	"FIRST",      "FL$DURATION",  "FL$STATUS",    "FOR",          "FREESESSION",  "FROM",       
	"GET",        "GETSESSION",   "GODIAL",       "GOFLOW",       "GOSUB",        "GOTO",       
	"HANGUP",     "HOSTRCV",      "HOSTSND",      "HT$SESSNUM",   "INDEX",        "IF",         
	"INPUT",      "K$ENTER",      "LEFTJUST",     "LENGTH",       "LENGTH",       "LET",        
	"LOWER",      "MAKECHAR",     "MAKENUM",      "MAKEPROMPT",   "MENU",         "MESSAGE",    
	"MOD",        "MONEY",        "MONTH",        "MSG$INFO",     "NEW",          "NEXT",       
	"NTH",        "NUM",          "NUMERIC",      "ON",           "PLACE",        "PROMPT",     
	"PUT",        "RANGE",        "RECORD",       "REMOVE",       "RESET",        "RETURN",     
	"RIGHTJUST",  "SAVE",         "SELECT",       "SEND",         "SPEAK",        "STARTSUB",   
	"STRING",     "SUBSTR",       "SYSPROMPT",    "SY$DAY",       "SY$MONTH",     "SY$YEAR",    
	"SY$HOUR",    "SY$MINUTE",    "SY$SECOND",    "SY$WEEKDAY",   "SY$DATE",      "SY$TIME",    
	"SY$RANDOM",  "SY$PORTINUSE", "THEN",         "TO",           "TRANSFER",     "TRANSLATE",  
	"UNLOCK",     "UPPER",        "USING",        "USR$GREET",    "USR$INFO",     "USR$NAME",   
	"WAIT",       "WITH",         "WRITEDB",     
	NIL,         
};


static long _cfw_in_comment( long offset );

void DLL _cfw_init( void );
BOOL DLL _cfw_expand( void );
void DLL _cfw_language_colors_update( long firstline, long lastline );
void DLL _cfw_language_colors( long numlines );
void DLL _cfw_assign_template( LPSTR abbrev, LPSTR template );
void DLL _cfw_add_keyword( LPSTR keyword );
BOOL DLL _cfw_indent( void );

void DLL
_init( void )
{
	_cfw_init();
}
/*
 ** _cfw_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_cfw_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "BOOL _cfw_expand");
		LibExport( "_cfw_assign_template LPSTR LPSTR");
		LibExport( "_cfw_add_keyword LPSTR");
		LibExport( "BOOL _cfw_indent");
		LibExport( "_cfw_language_colors_update long long");
		LibExport( "_cfw_language_colors long");

		firstTime = FALSE;

		/* Create a hash table to hold all of the 'CFW' keyword names.
		 */
		_cfw_hash_names = HashCreateTable( 539 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _cfw_names[ i ] != NIL )
		{
			LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _cfw_hash_names, 
					_cfw_names[i], strlen( _cfw_names[i] ) );
			if (lpf)
			{
				*lpf = StrNew( _cfw_names[i] );
				i++;
			}
		}

		/* Create a hash table to hold all of the 'CFW' template names.
		 */
		_cfw_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _cfw_templates[ i ].abbrev != NIL )
		{
			_cfw_assign_template( _cfw_templates[i].abbrev, _cfw_templates[i].template );
			i++;
		}
	}
}


/*
 ** _cfw_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'CFW' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_cfw_expand( void )
{
	LPVOID htEntry;
	LPSTR  _cfw_string;
	LPSTR  str;

	MarkSavePos();

	/* Check to see if we are currently positioned inside a 'CFW' comment
	 */
	if (_cfw_in_comment( -1L ) != -1)
		goto _cfw_expand_incomplete;

	str = BufReadStr( 65535 );     /* read entire line */

	_cfw_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_cfw_string)
		goto _cfw_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _cfw_hash_templates, _cfw_string, strlen( _cfw_string ) );
	StrFree( _cfw_string );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelToEOL();

		_ext_expand_template( str );

	} 
	else
		goto _cfw_expand_incomplete;
	return TRUE;

_cfw_expand_incomplete:
	MarkRestorePos();

	return FALSE;
}

/*
 ** _cfw_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
void DLL
_cfw_language_colors_update( long firstline, long lastline )
{
	long startoff;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);
		startoff = _PosQOffset();

		/* Clear all attributes for the line range and
		 * the reset them within _cfw_language_colors.
		 */
		AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		_PosInit( startoff );
		_cfw_language_colors( lastline - firstline + 1 );
	}
}

/*
 ** _cfw_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_cfw_language_colors( long numlines )
{
	long  startoff;    
	long  startline;   
	long  endoff;      
	long  inc;         
	WORD  ch;          
	LPSTR buf;         
	int   i             = 0;              
	int   comment_color = ColorComments(   -1 );
	int   keyword_color = ColorKeywords(   -1 );
	int   quote_color   = ColorAlternate1( -1 );
	int   number_color  = ColorAlternate2( -1 );

	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _cfw_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();


	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')
			numlines--;

		if (number_color && isdigit( ch ))
		{
            int j = 0;
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			startoff = _PosQOffset();
			do
			{
                j++;
				ch = _PosNextChar();
			} while (isxdigit( ch ) || (ch == '.') || (ch == 'x' && j == 1));

			if ((  lastch != ' ' 
				&& lastch != '\t' 
				&& lastch != '\r' 
				&& lastch != '\n' 
				&& lastch != EOF_CHAR
				&& !ispunct( lastch ))
				|| lastch == '_'
				|| (ch == '_'))
				continue;

            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );

		}
		else if (isalpha( (BYTE)ch ) || (ch == '_') || (ch == '.'))
		{
			WORD lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			i = 0;
			startoff = _PosQOffset();

			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char)ch;
				ch = _PosNextChar();
			} while (isalnum( (BYTE)ch ) || (ch == '_') || (ch == '.') || (ch == '$'));

			if (strchr( " \t\r\n[(:", lastch ) == NULL)
				i = 0;

			continue;
		}
		else if (i)
		{
			if (isalpha( (BYTE)buf[0] ))
			{
				if (HashFindEntry( _cfw_hash_names, buf, i ))
				{
					endoff = _PosQOffset();
					_PosInit( startoff );
					_PosSetColor( keyword_color, endoff - startoff );
					_PosInit( endoff );
				}
			}
			i = 0;
		} 

		if (ch == '"')
		{
			startoff = _PosQOffset();

			inc = _PosNextLine( 1 );
			numlines--;
			endoff = _PosQOffset();
			_PosInit( startoff );
			_PosSetColor( comment_color, endoff - startoff - inc );
			_PosInit( endoff );
			ch = _PosCurrentChar();
			continue;
		}
		ch = _PosNextChar();
	}
	MemFree( buf );
}


void DLL
_cfw_add_keyword( LPSTR keyword )
{
	if (keyword && *keyword)
	{
		LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _cfw_hash_names, 
								keyword, strlen( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}



/*
 ** _cfw_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'CFW' template strings
 *    without changing the code.
 */
void DLL
_cfw_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _cfw_hash_templates, abbrev, template );
}


/*
 ** _cfw_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'CFW' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in CFW_INDENT_STR - perform smart
 *       indentation.  
 */
BOOL DLL
_cfw_indent( void )
{
	long   curline;
	long   col;
	LPSTR  _cfw_string;
	LPSTR  str;
	long   matchpos;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'CFW' comment
	 */
	if (_cfw_in_comment( -1L ) != -1)
		goto _cfw_indent_incomplete;

	/* Not in a comment so indent according to the current 'CFW' construct.
	 */
	if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
			|| (BufQCurrentLine() != curline) )
		goto _cfw_indent_incomplete;

	/* Read the keyword at the beginning of the current line.
	 */
	col = BufQCurrentCol();
	SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
	str = BufReadStr( col - BufQCurrentCol() );
	_cfw_string = StrTrim( str, " \t" );
	StrFree( str );

	/* Set prefix and suffix to \377 to perform a quick lookup of the name
	 */
	str = StrNew( "\377" );
	_cfw_string = StrAppend( _cfw_string, str );
	str = StrAppend( str, _cfw_string );
	StrFree( _cfw_string );
	
	matchpos = StrMatch( str, CFW_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
	StrFree( str );

	if (!matchpos)
		/* Not a 'CFW' keyword, do normal indentation. */
		goto _cfw_indent_incomplete;
		
	MarkRestorePos();

	_ext_insert_indented_EOL();
	BufInsertChar( '\t' );
	return TRUE;

	_cfw_indent_incomplete:
		MarkRestorePos();

	return FALSE;
}



/*--  Private functions --*/


/*
 ** _cfw_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'CFW' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_cfw_in_comment( long offset )
{
	WORD  ch;

	/* Check to see if we are currently positioned inside a 'CFW' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		if (ch == '\n')
			return -1;
		else if (ch == '"')
			return _PosQOffset();  /* currrent line is in a comment */
		ch = _PosPrevChar();
	} while (ch != EOF_CHAR);

	return -1L;
}

