/* $Header:   E:/new/prgerr.c_v   1.0   10 Jun 1992 12:57:12   ericj  $ */
/*
 *
 *                     Copyright 1991 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "exports.h"
#include "startup.h"


/*
 ** _Clipper87ErrorInfo
 *
 *  This is the error info function to process error generated by
 *  the Clipper Summer'87 compiler.
 *
 *  The format of the error is:
 *    Compiling <filename>            <- only once per file
 *    "line <line>:  <message>"
 *    <line containing error>
 *    <'^' column position of error>
 *
 */
int DLL
_Clipper87ErrorInfo( void )
{
	LPSTR srchStr = "^(Compiling\\c|line \\c[0-9]+:)";
	LONG  pos;
	LPSTR msg;
	LPSTR srcName;

	/* Search through the error file for the next matching error message
	 * leaving the cursor located over the error's line number.
	 */
_Clipper87_search:
	if (!SrchFind( srchStr, SEARCH_FORWARD | SEARCH_REGEX | SEARCH_IGCASE, NIL ))
		return 0;
		
	/* Save the position of the line number and make a copy of the 
	 * message line.
	 */
	MarkSavePos();
	MovHome();
	msg = BufReadStr( 0x7FFFFFFF );
	MarkRestorePos();

	if ((msg[0] == 'C') || (msg[0] == 'c'))
	{
		/* Strip out the source file name from the error line.
		 */
		StrFree( msg );
		msg = BufReadStr( 0x7FFFF );
		srcName = StrTrim( msg, " " );
		StrFree( msg );
		msg = StrLtrim( srcName, " " );
		StrFree( srcName );
		ErrorSrcFile( msg );
		StrFree( msg );
		MovEOL();
		goto _Clipper87_search;
	}

	/* Determine the error message text.  In this case, everything
	 * past the ':'.
	 */
	pos = 0;
	while (msg[pos])
		if (msg[pos++] == ':')
			break;
	ErrorMsgLine( msg + pos );
	StrFree( msg );

	/* The current position is on the error line number.  Read the
	 * text into a string and convert it to a long integer.
	 */
	ErrorLine( atol( msg = BufReadStr( 12 )) );
	StrFree( msg );
	ErrorColumn( 0L );

	return 2;  /* always Error */
}

/*
 ** _Clipper5ErrorInfo
 *
 *  This is the error info function to process error generated by
 *  the Clipper 5.0 compiler.
 *
 *  The format of the error is:
 *    <sourcefile>(<line number>) "Error"|"Warning" <errno>  <message>
 */
int DLL
_Clipper5ErrorInfo( void )
{
	LPSTR srchStr = "\\(\\c[0-9]+\\)[ \"]+(Error|Warning)";
	int   level;
	LONG  pos;
	LONG  len;
	LPSTR msg;
	LPSTR srcName;
	WORD  ch;

	/* Search through the error file for the next matching error message
	 * leaving the cursor located over the error's line number.
	 */
	if (!SrchFind( srchStr, SEARCH_FORWARD | SEARCH_REGEX | SEARCH_IGCASE, NIL ))
		return 0;
		
	/* Save the position of the line number and make a copy of the 
	 * message line.
	 */
	MarkSavePos();
	MovHome();
	msg = BufReadStr( 0x7FFFFFFF );
	MarkRestorePos();

	/* Strip out the source file name from the error line.
	 */
	pos = StrMatch( "^ *\\c[^ ]+\\(", msg, 
		SEARCH_FORWARD | SEARCH_REGEX | SEARCH_MAXIMAL, &len );
	ErrorSrcFile( srcName = StrSubStr( msg, (int)pos, (int)len - 1 ));
	StrFree( srcName );
	StrFree( msg );

	/* The current position is on the error line number.  Read the
	 * text into a string and convert it to a long integer.
	 */
	ErrorLine( atol( msg = BufReadStr( 12 )) );
	StrFree( msg );
	ErrorColumn( 0L );

	/* Determine the severity of the error.
	 *
	 * return level == 1 for warning
	 *                 2 for error
	 */
	SrchFind( "Error|Warning", SEARCH_FORWARD | SEARCH_REGEX | SEARCH_IGCASE, &len );
	ch = BufReadChar();
	level = ((ch == 'e' || ch == 'E') ? 2 : 1);
	MovNextChar( len );

	/* Determine the error message text.  In this case, everything
	 * past the '(Erorr|Warning)'.
	 */
	msg = BufReadStr( 0x7FFFFFFF );  /* read 'til end of line */
	ErrorMsgLine( msg );
	StrFree( msg );

	return level;
}


